from abc import ABC, abstractmethod
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type, TypeVar

from django.conf import settings
from django.http import HttpRequest
from pydantic import BaseModel
from pydantic.fields import FieldInfo

from ninja.compatibility import get_headers
from ninja.errors import HttpError
from ninja.types import DictStrAny

if TYPE_CHECKING:
    from ninja import NinjaAPI  # pragma: no cover

__all__ = [
    "ParamModel",
    "QueryModel",
    "PathModel",
    "HeaderModel",
    "CookieModel",
    "BodyModel",
    "FormModel",
    "FileModel",
]

TModel = TypeVar("TModel", bound="ParamModel")
TModels = List[TModel]


def NestedDict() -> DictStrAny:
    return defaultdict(NestedDict)


class ParamModel(BaseModel, ABC):
    __ninja_param_source__ = None

    @classmethod
    @abstractmethod
    def get_request_data(
        cls, request: HttpRequest, api: "NinjaAPI", path_params: DictStrAny
    ) -> Optional[DictStrAny]:
        pass  # pragma: no cover

    @classmethod
    def resolve(
        cls: Type[TModel],
        request: HttpRequest,
        api: "NinjaAPI",
        path_params: DictStrAny,
    ) -> TModel:
        data = cls.get_request_data(request, api, path_params)
        if data is None:
            return cls()

        data = cls._map_data_paths(data)
        return cls.model_validate(data, context={"request": request})

    @classmethod
    def _map_data_paths(cls, data: DictStrAny) -> DictStrAny:
        flatten_map = getattr(cls, "__ninja_flatten_map__", None)
        if not flatten_map:
            return data

        mapped_data: DictStrAny = NestedDict()
        for k in flatten_map:
            if k in data:
                cls._map_data_path(mapped_data, data[k], flatten_map[k])
            else:
                cls._map_data_path(mapped_data, None, flatten_map[k])

        return mapped_data

    @classmethod
    def _map_data_path(cls, data: DictStrAny, value: Any, path: Tuple) -> None:
        if len(path) == 1:
            if value is not None:
                data[path[0]] = value
        else:
            cls._map_data_path(data[path[0]], value, path[1:])


class QueryModel(ParamModel):
    @classmethod
    def get_request_data(
        cls, request: HttpRequest, api: "NinjaAPI", path_params: DictStrAny
    ) -> Optional[DictStrAny]:
        list_fields = getattr(cls, "__ninja_collection_fields__", [])
        return api.parser.parse_querydict(request.GET, list_fields, request)


class PathModel(ParamModel):
    @classmethod
    def get_request_data(
        cls, request: HttpRequest, api: "NinjaAPI", path_params: DictStrAny
    ) -> Optional[DictStrAny]:
        return path_params


class HeaderModel(ParamModel):
    __ninja_flatten_map__: DictStrAny

    @classmethod
    def get_request_data(
        cls, request: HttpRequest, api: "NinjaAPI", path_params: DictStrAny
    ) -> Optional[DictStrAny]:
        data = {}
        headers = get_headers(request)
        for name in cls.__ninja_flatten_map__:
            if name in headers:
                data[name] = headers[name]
        return data


class CookieModel(ParamModel):
    @classmethod
    def get_request_data(
        cls, request: HttpRequest, api: "NinjaAPI", path_params: DictStrAny
    ) -> Optional[DictStrAny]:
        return request.COOKIES


class BodyModel(ParamModel):
    __read_from_single_attr__: str

    @classmethod
    def get_request_data(
        cls, request: HttpRequest, api: "NinjaAPI", path_params: DictStrAny
    ) -> Optional[DictStrAny]:
        if request.body:
            try:
                data = api.parser.parse_body(request)
            except Exception as e:
                msg = "Cannot parse request body"
                if settings.DEBUG:
                    msg += f" ({e})"
                raise HttpError(400, msg) from e

            varname = getattr(cls, "__read_from_single_attr__", None)
            if varname:
                data = {varname: data}
            return data

        return None


class FormModel(ParamModel):
    @classmethod
    def get_request_data(
        cls, request: HttpRequest, api: "NinjaAPI", path_params: DictStrAny
    ) -> Optional[DictStrAny]:
        list_fields = getattr(cls, "__ninja_collection_fields__", [])
        return api.parser.parse_querydict(request.POST, list_fields, request)


class FileModel(ParamModel):
    @classmethod
    def get_request_data(
        cls, request: HttpRequest, api: "NinjaAPI", path_params: DictStrAny
    ) -> Optional[DictStrAny]:
        list_fields = getattr(cls, "__ninja_collection_fields__", [])
        return api.parser.parse_querydict(request.FILES, list_fields, request)


class _HttpRequest(HttpRequest):
    body: bytes = b""


class _MultiPartBodyModel(BodyModel):
    __ninja_body_params__: DictStrAny

    @classmethod
    def get_request_data(
        cls, request: HttpRequest, api: "NinjaAPI", path_params: DictStrAny
    ) -> Optional[DictStrAny]:
        req = _HttpRequest()
        get_request_data = super(_MultiPartBodyModel, cls).get_request_data
        results: DictStrAny = {}
        for name, annotation in cls.__ninja_body_params__.items():
            if name in request.POST:
                data = request.POST[name]
                if annotation == str and data[0] != '"' and data[-1] != '"':
                    data = f'"{data}"'
                req.body = data.encode()
                results[name] = get_request_data(req, api, path_params)
        return results


class Param(FieldInfo):
    def __init__(
        self,
        default: Any,
        *,
        alias: str = None,
        title: str = None,
        description: str = None,
        gt: float = None,
        ge: float = None,
        lt: float = None,
        le: float = None,
        min_length: int = None,
        max_length: int = None,
        regex: str = None,
        example: Any = None,
        examples: Optional[Dict[str, Any]] = None,
        deprecated: bool = None,
        include_in_schema: bool = True,
        # param_name: str = None,
        # param_type: Any = None,
        **extra: Any,
    ):
        self.deprecated = deprecated
        # self.param_name: str = None
        # self.param_type: Any = None
        self.model_field: Optional[FieldInfo] = None
        json_schema_extra = {}
        if example:
            json_schema_extra["example"] = example
        if examples:
            json_schema_extra["examples"] = examples
        if deprecated:
            json_schema_extra["deprecated"] = deprecated
        if not include_in_schema:
            json_schema_extra["include_in_schema"] = include_in_schema
        if alias and not extra.get("validation_alias"):
            extra["validation_alias"] = alias
        if alias and not extra.get("serialization_alias"):
            extra["serialization_alias"] = alias
        super().__init__(
            default=default,
            alias=alias,
            title=title,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            json_schema_extra=json_schema_extra,
            **extra,
        )

    @classmethod
    def _param_source(cls) -> str:
        "Openapi param.in value or body type"
        return cls.__name__.lower()


class Path(Param):
    _model = PathModel


class Query(Param):
    _model = QueryModel


class Header(Param):
    _model = HeaderModel


class Cookie(Param):
    _model = CookieModel


class Body(Param):
    _model = BodyModel


class Form(Param):
    _model = FormModel


class File(Param):
    _model = FileModel


class _MultiPartBody(Param):
    _model = _MultiPartBodyModel

    @classmethod
    def _param_source(cls) -> str:
        return "body"



"""
Django settings for multi_param project.

Generated by 'django-admin startproject' using Django 3.1.13.

For more information on this file, see
https://docs.djangoproject.com/en/3.1/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.1/ref/settings/
"""

from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.1/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "NOT-SUPER-SECRET-DO-NOT-USE-ME"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []

# Application definition

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "tests.demo_project.multi_param.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "tests.demo_project.multi_param.wsgi.application"
DATABASES = {"default": {}}
AUTH_PASSWORD_VALIDATORS = []

LANGUAGE_CODE = "en-us"
TIME_ZONE = "UTC"
USE_I18N = True
USE_TZ = False
STATIC_URL = "/static/"



from ninja import NinjaAPI
from ninja.security import HttpBearer

api = NinjaAPI()

class InvalidToken(Exception):
    pass

@api.exception_handler(InvalidToken)
def on_invalid_token(request, exc):
    return api.create_response(request, {"detail": "Invalid token supplied"}, status=401)

class AuthBearer(HttpBearer):
    def authenticate(self, request, token):
        if token == "supersecret":
            return token
        raise InvalidToken


@api.get("/bearer", auth=AuthBearer())
def bearer(request):
    return {"token": request.auth}



import pytest

from ninja import Form, NinjaAPI, Schema
from ninja.errors import ConfigError
from ninja.testing import TestClient

api = NinjaAPI()


@api.post("/form")
def form_operation(request, s: str = Form(...), i: int = Form(None)):
    return {"s": s, "i": i}


client = TestClient(api)


def test_form():
    response = client.post("/form")  # invalid
    assert response.status_code == 422

    response = client.post("/form", POST={"s": "text"})
    assert response.status_code == 200
    assert response.json() == {"i": None, "s": "text"}

    response = client.post("/form", POST={"s": "text", "i": None})
    assert response.status_code == 200
    assert response.json() == {"i": None, "s": "text"}

    response = client.post("/form", POST={"s": "text", "i": 2})
    assert response.status_code == 200
    assert response.json() == {"i": 2, "s": "text"}


def test_schema():
    schema = api.get_openapi_schema()
    method = schema["paths"]["/api/form"]["post"]
    print(method["requestBody"])
    assert method["requestBody"] == {
        "content": {
            "application/x-www-form-urlencoded": {
                "schema": {
                    "properties": {
                        "i": {"type": "integer", "title": "I"},
                        "s": {"type": "string", "title": "S"},
                    },
                    "required": ["s"],
                    "title": "FormParams",
                    "type": "object",
                }
            }
        },
        "required": True,
    }


def test_duplicate_names():
    class TestData(Schema):
        p1: str

    match = "Duplicated name: 'p1' in params: 'p1' & 'data'"
    with pytest.raises(ConfigError, match=match):

        @api.post("/broken1")
        def broken1(request, p1: int = Form(...), data: TestData = Form(...)):
            pass

    match = "Duplicated name: 'p1' also in 'data'"
    with pytest.raises(ConfigError, match=match):

        @api.post("/broken2")
        def broken2(request, data: TestData = Form(...), p1: int = Form(...)):
            pass


# TODO: Fix schema for this case:
# class Credentials(Schema):
#     username: str
#     password: str


# @api.post("/login")
# def login(request, credentials: Credentials = Form(...)):
#     return {'username': credentials.username}



import json
from pathlib import Path

import pytest
from django.core.files.uploadedfile import SimpleUploadedFile
from django.test import Client as DjangoTestClient

from ninja.testing import TestClient as NinjaTestClient
from tests.demo_project.multi_param.api import router

ninja_client = NinjaTestClient(router)

test_file = SimpleUploadedFile("test.txt", b"data123")

expected_response = {
    "i": 1,
    "s": "a-str",
    "data": {"foo4": 44, "bar4": "44bar"},
    "nested-data": {
        "foo": 11,
        "bar": "11bar",
        "d2": {"foo2": 22, "bar2": "22bar", "d3": {"foo3": 33, "bar3": "33bar"}},
    },
}

test_client_args = {
    "/test-multi-query": {
        "query": "i=1&s=a-str&foo4=44&bar4=44bar&foo=11&bar=11bar&foo2=22&bar2=22bar&foo3=33&bar3=33bar",
    },
    "/test-multi-path/{i}/{s}/{foo4}/{bar4}/{foo}/{bar}/{foo2}/{bar2}/{foo3}/{bar3}/": {
        "path": "/test-multi-path/1/a-str/44/44bar/11/11bar/22/22bar/33/33bar/"
    },
    "/test-multi-header": {"headers": {"i": 1, "foo": 11, "bar2": "22bar", "foo3": 33}},
    "/test-multi-cookie": {"COOKIES": {"i": 1, "foo": 11, "bar2": "22bar", "foo3": 33}},
    "/test-multi-form": {"POST": {"i": 1, "foo": 11, "bar2": "22bar", "foo3": 33}},
    "/test-multi-body": {"json": expected_response},
    "/test-multi-body-file": {
        "FILES": {"file": test_file},
        "POST": {
            k: json.dumps(v) if isinstance(v, dict) else str(v)
            for k, v in expected_response.items()
        },
    },
    "/test-multi-form-file": {
        "FILES": {"file": test_file},
        "POST": {"i": 1, "foo": 11, "bar2": "22bar", "foo3": 33},
    },
    "/test-multi-body-form": {
        "POST": {
            "i": "1",
            "foo": 11,
            "bar2": "22bar",
            "foo3": 33,
            "data": '{"foo4": 44, "bar4": "44bar"}',
        }
    },
    "/test-multi-form-body": {
        "POST": {
            "i": 1,
            "nested-data": '{"foo": 11, "d2": {"bar2": "22bar", "d3": {"foo3": 33}}}',
        }
    },
    "/test-multi-body-form-file": {
        "FILES": {"file": test_file},
        "POST": {
            "i": "1",
            "foo": 11,
            "bar2": "22bar",
            "foo3": 33,
            "data": '{"foo4": 44, "bar4": "44bar"}',
        },
    },
    "/test-multi-form-body-file": {
        "FILES": {"file": test_file},
        "POST": {
            "i": 1,
            "nested-data": '{"foo": 11, "d2": {"bar2": "22bar", "d3": {"foo3": 33}}}',
        },
    },
}


def test_validate_test_data():
    ops = router.path_operations
    schema = DjangoTestClient().get("/api/mp/openapi.json").json()

    for path in tuple(schema["paths"]):
        schema["paths"]["/" + path.split("/", 3)[3]] = schema["paths"].pop(path)
    assert set(ops) == set(
        schema["paths"]
    ), "Expect a test case for each endpoint on the API"

    fixture_dir = Path(__file__).parent / "schema_fixtures"
    fixture_files = {
        path: (fixture_dir / f"{path.split('/', 2)[1]}.json") for path in ops
    }

    # verify that the currently generated schema matches the fixtures.  Since the generated
    # schema is in git, this allows any changes in behavior to be documented over time.
    for path, filename in fixture_files.items():
        if 0:  # pragma: nocover
            # if test cases or schema generation changes,
            #   use this block of code to regenerate the fixtures
            with open(filename, "w") as f:
                json.dump(schema["paths"][path], f, indent=2)
        with open(filename) as f:
            data = json.load(f)
            assert json.loads(json.dumps(schema["paths"][path])) == data


@pytest.mark.parametrize("path, client_args", tuple(test_client_args.items()))
def test_data_round_trip_with_ninja_client(path, client_args):
    client_args = test_client_args[path]
    kwargs = {"path": path}
    for k in ("path", "headers", "COOKIES", "POST", "json", "FILES"):
        if k in client_args:
            kwargs[k] = client_args[k]

    query = client_args.get("query")
    if query:
        kwargs["path"] += f"?{query}"

    response = ninja_client.post(**kwargs)
    assert response.json() == expected_response
    assert response.status_code == 200


@pytest.mark.parametrize("path, client_args", tuple(test_client_args.items()))
def test_data_round_trip_with_django_client(path, client_args):
    django_client = DjangoTestClient()

    client_args = test_client_args[path]
    kwargs = {"path": client_args.get("path", path)}

    if "headers" in client_args:
        for k, v in client_args["headers"].items():
            kwargs[f"HTTP_{k.upper()}"] = v

    if "COOKIES" in client_args:
        django_client.cookies.load(client_args["COOKIES"])

    if "POST" in client_args:
        kwargs["data"] = client_args["POST"]

    if "FILES" in client_args:
        if "data" in kwargs:
            kwargs["data"].update(client_args["FILES"])
        else:
            kwargs["data"] = client_args["FILES"]

    if "json" in client_args:
        assert "data" not in kwargs
        kwargs["data"] = json.dumps(client_args["json"])
        kwargs["content_type"] = "application/json"

    query = client_args.get("query")
    if query:
        kwargs["path"] += f"?{query}"

    kwargs["path"] = "/api/mp" + kwargs["path"]

    response = django_client.post(**kwargs)
    assert response.json() == expected_response
    assert response.status_code == 200



from abc import ABC, abstractmethod
from typing import Any, Optional

from django.http import HttpRequest

from ninja.errors import ConfigError
from ninja.utils import is_async_callable

__all__ = ["SecuritySchema", "AuthBase"]


class SecuritySchema(dict):
    def __init__(self, type: str, **kwargs: Any) -> None:
        super().__init__(type=type, **kwargs)


class AuthBase(ABC):
    def __init__(self) -> None:
        if not hasattr(self, "openapi_type"):
            raise ConfigError("If you extend AuthBase you need to define openapi_type")

        kwargs = {}
        for attr in dir(self):
            if attr.startswith("openapi_"):
                name = attr.replace("openapi_", "", 1)
                kwargs[name] = getattr(self, attr)
        self.openapi_security_schema = SecuritySchema(**kwargs)

        self.is_async = False
        if hasattr(self, "authenticate"):  # pragma: no branch
            self.is_async = is_async_callable(self.authenticate)  # type: ignore

    @abstractmethod
    def __call__(self, request: HttpRequest) -> Optional[Any]:
        pass  # pragma: no cover



import json
import os
import tempfile
from io import StringIO
from unittest.mock import patch

import pytest
from django.core.management import call_command
from django.core.management.base import CommandError

from ninja.management.commands.export_openapi_schema import Command as ExportCmd


def test_export_default():
    output = StringIO()
    call_command(ExportCmd(), stdout=output)
    json.loads(output.getvalue())  # if no exception, then OK
    assert len(output.getvalue().splitlines()) == 1


def test_export_indent():
    output = StringIO()
    call_command(ExportCmd(), indent=1, stdout=output)
    assert len(output.getvalue().splitlines()) > 1


def test_export_to_file():
    with tempfile.TemporaryDirectory() as tmp:
        output_file = os.path.join(tmp, "result.json")
        call_command(ExportCmd(), output=output_file)
        with open(output_file, "r") as f:
            json.loads(f.read())


def test_export_custom():
    with pytest.raises(CommandError):
        call_command(ExportCmd(), api="something.that.doesnotexist")

    with pytest.raises(CommandError) as e:
        call_command(ExportCmd(), api="django.core.management.base.BaseCommand")
    assert (
        str(e.value)
        == "django.core.management.base.BaseCommand is not instance of NinjaAPI!"
    )

    call_command(ExportCmd(), api="demo.urls.api_v1")
    call_command(ExportCmd(), api="demo.urls.api_v2")


@patch("ninja.management.commands.export_openapi_schema.resolve")
def test_export_default_without_api_endpoint(mock):
    mock.side_effect = AttributeError()
    output = StringIO()
    with pytest.raises(CommandError) as e:
        call_command(ExportCmd(), stdout=output)
    assert str(e.value) == "No NinjaAPI instance found; please specify one with --api"



from ninja import NinjaAPI, Router
from ninja.security import APIKeyQuery
from ninja.testing import TestClient


class KeyQuery1(APIKeyQuery):
    def authenticate(self, request, key):
        if key == "k1":
            return key


class KeyQuery2(APIKeyQuery):
    def authenticate(self, request, key):
        if key == "k2":
            return key


api = NinjaAPI(auth=KeyQuery1())


@api.get("/default")
def default(request):
    return {"auth": request.auth}


@api.api_operation(["POST", "PATCH"], "/multi-method-no-auth")
def multi_no_auth(request):
    return {"auth": request.auth}


@api.api_operation(["POST", "PATCH"], "/multi-method-auth", auth=KeyQuery2())
def multi_auth(request):
    return {"auth": request.auth}


# ---- router ------------------------

router = Router()


@router.get("/router-operation")  # should come from global auth
def router_operation(request):
    return {"auth": str(request.auth)}


@router.get("/router-operation-auth", auth=KeyQuery2())
def router_operation_auth(request):
    return {"auth": str(request.auth)}


api.add_router("", router)

# ---- end router --------------------

client = TestClient(api)


def test_multi():
    assert client.get("/default").status_code == 401
    assert client.get("/default?key=k1").json() == {"auth": "k1"}

    assert client.post("/multi-method-no-auth").status_code == 401
    assert client.post("/multi-method-no-auth?key=k1").json() == {"auth": "k1"}

    assert client.patch("/multi-method-no-auth").status_code == 401
    assert client.patch("/multi-method-no-auth?key=k1").json() == {"auth": "k1"}

    assert client.post("/multi-method-auth?key=k1").status_code == 401
    assert client.patch("/multi-method-auth?key=k1").status_code == 401

    assert client.post("/multi-method-auth?key=k2").json() == {"auth": "k2"}
    assert client.patch("/multi-method-auth?key=k2").json() == {"auth": "k2"}


def test_router_auth():
    assert client.get("/router-operation").status_code == 401
    assert client.get("/router-operation?key=k1").json() == {"auth": "k1"}

    assert client.get("/router-operation-auth?key=k1").status_code == 401
    assert client.get("/router-operation-auth?key=k2").json() == {"auth": "k2"}



from datetime import date
from enum import Enum
from typing import List, Optional

from pydantic import BaseModel

from ninja import NinjaAPI, Query
from ninja.testing import TestClient


class RoomEnum(str, Enum):
    double = "double"
    twin = "twin"
    single = "single"


class ExtraEnum(str, Enum):
    a = "a"
    b = "b"


class Booking(BaseModel):
    start: date
    end: date
    room: RoomEnum = RoomEnum.double


api = NinjaAPI()


@api.post("/book")
def create_booking(request, booking: Booking):
    return booking


@api.get("/search")
def booking_search(request, room: RoomEnum):
    return {"room": room}


@api.get("/optional")
def enum_optional(
    request, room: Optional[RoomEnum] = Query(None, description="description")
):
    return {"room": room}


@api.get("/optional2")
def enum_optional2(request, extra: Optional[ExtraEnum] = None):
    return {"extra": extra}


@api.get("/list")
def enum_list(request, rooms: List[RoomEnum] = Query(None, description="description")):
    return {"rooms": rooms}


class QueryOnlyEnum(str, Enum):
    one = "one"
    two = "two"


@api.get("/new-list")
def new_enum_list(
    request, q: List[QueryOnlyEnum] = Query(None, description="description")
):
    return {"q": q}


client = TestClient(api)


def test_enums():
    response = client.post(
        "/book", json={"start": "2020-01-01", "end": "2020-01-02", "room": "double"}
    )
    assert response.status_code == 200, response.content
    assert response.json() == {
        "start": "2020-01-01",
        "end": "2020-01-02",
        "room": "double",
    }

    response = client.post(
        "/book", json={"start": "2020-01-01", "end": "2020-01-02", "room": "triple"}
    )
    assert response.status_code == 422

    response = client.get("/search?room=twin")
    assert response.status_code == 200
    assert response.json() == {"room": "twin"}

    response = client.get("/search?room=other")
    assert response.status_code == 422

    response = client.get("/optional?room=twin")
    assert response.status_code == 200

    response = client.get("/optional")
    assert response.status_code == 200
    assert response.json() == {"room": None}

    response = client.get("/optional2?extra=a")
    assert response.status_code == 200
    assert response.json() == {"extra": "a"}

    response = client.get("/optional2")
    assert response.json() == {"extra": None}

    response = client.get("/list?rooms=twin&rooms=single")
    assert response.status_code == 200
    assert response.json() == {"rooms": ["twin", "single"]}

    response = client.get("/new-list?q=one&q=one")
    assert response.status_code == 200
    assert response.json() == {"q": ["one", "one"]}


def test_schema():
    schema = api.get_openapi_schema()

    booking_schema = schema["components"]["schemas"]["Booking"]
    room_prop = booking_schema["properties"]["room"]

    if "allOf" in room_prop:
        # pydantic 1.7+ change:
        assert room_prop["allOf"] == [{"$ref": "#/components/schemas/RoomEnum"}]
    else:
        assert room_prop == {"$ref": "#/components/schemas/RoomEnum"}

    assert schema["components"]["schemas"]["RoomEnum"] == {
        "enum": ["double", "twin", "single"],
        "title": "RoomEnum",
        "type": "string",
    }

    book_operation = schema["paths"]["/api/book"]["post"]
    assert book_operation["requestBody"]["content"]["application/json"]["schema"] == {
        "$ref": "#/components/schemas/Booking"
    }

    search_operation = schema["paths"]["/api/search"]["get"]
    room_param = search_operation["parameters"][0]
    assert room_param == {
        "in": "query",
        "name": "room",
        "required": True,
        "schema": {
            "title": "RoomEnum",
            "enum": ["double", "twin", "single"],
            "type": "string",
        },
    }

    optional_operation = schema["paths"]["/api/optional"]["get"]
    room_param = optional_operation["parameters"][0]
    assert room_param == {
        "in": "query",
        "name": "room",
        "schema": {
            "anyOf": [{"$ref": "#/components/schemas/RoomEnum"}, {"type": "null"}],
            "description": "description",
        },
        "required": False,
        "description": "description",
    }

    assert schema["paths"]["/api/new-list"]["get"]["parameters"][0] == {
        "description": "description",
        "in": "query",
        "name": "q",
        "required": False,
        "schema": {
            "description": "description",
            "title": "Q",
            "items": {
                "enum": ["one", "two"],
                "title": "QueryOnlyEnum",
                "type": "string",
            },
            "type": "array",
        },
    }


def test_optional_get_schema():
    "This tests that enum that is only used in GET operation puts a that enum into schema.components"
    schema = api.get_openapi_schema()

    op = schema["paths"]["/api/optional2"]["get"]
    print(op)
    assert op["parameters"][0]["schema"]["anyOf"] == [
        {"$ref": "#/components/schemas/ExtraEnum"},
        {"type": "null"},
    ]

    components = schema["components"]["schemas"]
    print(components)
    assert "ExtraEnum" in components



import logging
from abc import ABC, abstractmethod
from base64 import b64decode
from typing import Any, Optional, Tuple
from urllib.parse import unquote

from django.conf import settings
from django.http import HttpRequest

from ninja.compatibility import get_headers
from ninja.security.base import AuthBase

__all__ = ["HttpAuthBase", "HttpBearer", "DecodeError", "HttpBasicAuth"]


logger = logging.getLogger("django")


class HttpAuthBase(AuthBase, ABC):
    openapi_type: str = "http"


class HttpBearer(HttpAuthBase, ABC):
    openapi_scheme: str = "bearer"
    header: str = "Authorization"

    def __call__(self, request: HttpRequest) -> Optional[Any]:
        headers = get_headers(request)
        auth_value = headers.get(self.header)
        if not auth_value:
            return None
        parts = auth_value.split(" ")

        if parts[0].lower() != self.openapi_scheme:
            if settings.DEBUG:
                logger.error(f"Unexpected auth - '{auth_value}'")
            return None
        token = " ".join(parts[1:])
        return self.authenticate(request, token)

    @abstractmethod
    def authenticate(self, request: HttpRequest, token: str) -> Optional[Any]:
        pass  # pragma: no cover


class DecodeError(Exception):
    pass


class HttpBasicAuth(HttpAuthBase, ABC):  # TODO: maybe HttpBasicAuthBase
    openapi_scheme = "basic"
    header = "Authorization"

    def __call__(self, request: HttpRequest) -> Optional[Any]:
        headers = get_headers(request)
        auth_value = headers.get(self.header)
        if not auth_value:
            return None

        try:
            username, password = self.decode_authorization(auth_value)
        except DecodeError as e:
            if settings.DEBUG:
                logger.exception(e)
            return None
        return self.authenticate(request, username, password)

    @abstractmethod
    def authenticate(
        self, request: HttpRequest, username: str, password: str
    ) -> Optional[Any]:
        pass  # pragma: no cover

    def decode_authorization(self, value: str) -> Tuple[str, str]:
        parts = value.split(" ")
        if len(parts) == 1:
            user_pass_encoded = parts[0]
        elif len(parts) == 2 and parts[0].lower() == "basic":
            user_pass_encoded = parts[1]
        else:
            raise DecodeError("Invalid Authorization header")

        try:
            username, password = b64decode(user_pass_encoded).decode().split(":", 1)
            return unquote(username), unquote(password)
        except (
            Exception
        ) as e:  # dear contributors please do not change to valueerror - here can be multiple exceptions
            raise DecodeError("Invalid Authorization header") from e



import os
from tempfile import NamedTemporaryFile

import pytest
from django.http import FileResponse, HttpResponse

from ninja import NinjaAPI
from ninja.main import ConfigError
from ninja.testing import TestClient

api = NinjaAPI()

client = TestClient(api)

# TODO: check if you add  operation to the same path - it should raise a ConfigError that this path already exist
# make sure to check how this will work with versioning
# and also check what will happen if you add same path in different routers
#  api.add_router("", router1)
#  api.add_router("", router2)
# and both routers have same path defined


@api.get("")
def emptypath(request):
    return "/"


@api.get("/get")
def get(request):
    return f"this is {request.method}"


@api.post("/post")
def post(request):
    return f"this is {request.method}"


@api.put("/put")
def put(request):
    return f"this is {request.method}"


@api.patch("/patch")
def patch(request):
    return f"this is {request.method}"


@api.delete("/delete")
def delete(request):
    return f"this is {request.method}"


@api.api_operation(["GET", "POST"], "/multi")
def multiple(request):
    return f"this is {request.method}"


@api.get("/html")
def html(request):
    return HttpResponse("html")


@api.get("/file")
def file_response(request):
    tmp = NamedTemporaryFile(delete=False)
    try:
        with open(tmp.name, "wb") as f:
            f.write(b"this is a file")
        return FileResponse(open(tmp.name, "rb"))
    finally:
        try:
            os.remove(tmp.name)
        except PermissionError:
            pass


@pytest.mark.parametrize(
    # fmt: off
    "method,path,expected_status,expected_data,expected_streaming",
    [
        ("get",    "/",       200, "/", False),
        ("get",    "/get",    200, "this is GET", False),
        ("post",   "/post",   200, "this is POST", False),
        ("put",    "/put",    200, "this is PUT", False),
        ("patch",  "/patch",  200, "this is PATCH", False),
        ("delete", "/delete", 200, "this is DELETE", False),
        ("get",    "/multi",  200, "this is GET", False),
        ("post",   "/multi",  200, "this is POST", False),
        ("patch",  "/multi",  405, b"Method not allowed", False),
        ("get",    "/html",   200, b"html", False),
        ("get",    "/file",   200, b"this is a file", True),
    ],
    # fmt: on
)
def test_method(method, path, expected_status, expected_data, expected_streaming):
    func = getattr(client, method)
    response = func(path)
    assert response.status_code == expected_status
    assert response.streaming == expected_streaming
    try:
        data = response.json()
    except Exception:
        data = response.content
    assert data == expected_data


def test_validates():
    try:
        os.environ["NINJA_SKIP_REGISTRY"] = ""
        with pytest.raises(ConfigError):
            _urls = NinjaAPI().urls
    finally:
        os.environ["NINJA_SKIP_REGISTRY"] = "yes"



from ninja.security import APIKeyCookie


class CookieKey(APIKeyCookie):
    def authenticate(self, request, key):
        if key == "supersecret":
            return key


cookie_key = CookieKey()


@api.get("/cookiekey", auth=cookie_key)
def apikey(request):
    return f"Token = {request.auth}"



from django.http import HttpResponse

from ninja import NinjaAPI
from ninja.testing import TestClient

api = NinjaAPI()


@api.get("/test-no-cookies")
def op_no_cookies(request):
    return {}


@api.get("/test-set-cookie")
def op_set_cookie(request):
    response = HttpResponse()
    response.set_cookie(key="sessionid", value="sessionvalue")
    return response


client = TestClient(api)


def test_cookies():
    assert bool(client.get("/test-no-cookies").cookies) is False
    assert "sessionid" in client.get("/test-set-cookie").cookies



import inspect
from abc import ABC, abstractmethod
from functools import partial, wraps
from typing import Any, Callable, List, Optional, Tuple, Type

from django.db.models import QuerySet
from django.http import HttpRequest
from django.utils.module_loading import import_string
from typing_extensions import get_args as get_collection_args  # type: ignore

from ninja import Field, Query, Router, Schema
from ninja.conf import settings
from ninja.constants import NOT_SET
from ninja.errors import ConfigError
from ninja.operation import Operation
from ninja.signature.details import is_collection_type
from ninja.utils import contribute_operation_args, contribute_operation_callback


class PaginationBase(ABC):
    class Input(Schema):
        pass

    InputSource = Query(...)

    class Output(Schema):
        items: List[Any]
        count: int

    items_attribute: str = "items"

    def __init__(self, *, pass_parameter: Optional[str] = None, **kwargs: Any) -> None:
        self.pass_parameter = pass_parameter

    @abstractmethod
    def paginate_queryset(
        self,
        queryset: QuerySet,
        pagination: Any,
        **params: Any,
    ) -> Any:
        pass  # pragma: no cover

    def _items_count(self, queryset: QuerySet) -> int:
        """
        Since lists are mainly compatible with QuerySets and can be passed to paginator.
        We will first to try to use .count - and if not there will use a len
        """
        try:
            # forcing to find queryset.count instead of list.count:
            return queryset.all().count()  # type: ignore
        except AttributeError:
            return len(queryset)


class LimitOffsetPagination(PaginationBase):
    class Input(Schema):
        limit: int = Field(settings.PAGINATION_PER_PAGE, ge=1)
        offset: int = Field(0, ge=0)

    def paginate_queryset(
        self,
        queryset: QuerySet,
        pagination: Input,
        **params: Any,
    ) -> Any:
        offset = pagination.offset
        limit: int = pagination.limit
        return {
            "items": queryset[offset : offset + limit],
            "count": self._items_count(queryset),
        }  # noqa: E203


class PageNumberPagination(PaginationBase):
    class Input(Schema):
        page: int = Field(1, ge=1)

    def __init__(
        self, page_size: int = settings.PAGINATION_PER_PAGE, **kwargs: Any
    ) -> None:
        self.page_size = page_size
        super().__init__(**kwargs)

    def paginate_queryset(
        self,
        queryset: QuerySet,
        pagination: Input,
        **params: Any,
    ) -> Any:
        offset = (pagination.page - 1) * self.page_size
        return {
            "items": queryset[offset : offset + self.page_size],
            "count": self._items_count(queryset),
        }  # noqa: E203


def paginate(func_or_pgn_class: Any = NOT_SET, **paginator_params: Any) -> Callable:
    """
    @api.get(...
    @paginate
    def my_view(request):

    or

    @api.get(...
    @paginate(PageNumberPagination)
    def my_view(request):

    """

    isfunction = inspect.isfunction(func_or_pgn_class)
    isnotset = func_or_pgn_class == NOT_SET

    pagination_class: Type[PaginationBase] = import_string(settings.PAGINATION_CLASS)

    if isfunction:
        return _inject_pagination(func_or_pgn_class, pagination_class)

    if not isnotset:
        pagination_class = func_or_pgn_class

    def wrapper(func: Callable) -> Any:
        return _inject_pagination(func, pagination_class, **paginator_params)

    return wrapper


def _inject_pagination(
    func: Callable,
    paginator_class: Type[PaginationBase],
    **paginator_params: Any,
) -> Callable:
    paginator: PaginationBase = paginator_class(**paginator_params)

    @wraps(func)
    def view_with_pagination(request: HttpRequest, **kwargs: Any) -> Any:
        pagination_params = kwargs.pop("ninja_pagination")
        if paginator.pass_parameter:
            kwargs[paginator.pass_parameter] = pagination_params

        items = func(request, **kwargs)

        result = paginator.paginate_queryset(
            items, pagination=pagination_params, request=request, **kwargs
        )
        if paginator.Output:
            result[paginator.items_attribute] = list(result[paginator.items_attribute])
            # ^ forcing queryset evaluation #TODO: check why pydantic did not do it here
        return result

    contribute_operation_args(
        view_with_pagination,
        "ninja_pagination",
        paginator.Input,
        paginator.InputSource,
    )

    if paginator.Output:
        contribute_operation_callback(
            view_with_pagination,
            partial(make_response_paginated, paginator),
        )

    return view_with_pagination


class RouterPaginated(Router):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.pagination_class = import_string(settings.PAGINATION_CLASS)

    def add_api_operation(
        self, path: str, methods: List[str], view_func: Callable, **kwargs: Any
    ) -> None:
        response = kwargs["response"]
        if is_collection_type(response):
            view_func = _inject_pagination(view_func, self.pagination_class)
        return super().add_api_operation(path, methods, view_func, **kwargs)


def make_response_paginated(paginator: PaginationBase, op: Operation) -> None:
    """
    Takes operation response and changes it to the paginated response
    for example:
        response=List[Some]
    will be changed to:
        response=PagedSome
    where Paged some will be a subclass of paginator.Output:
        class PagedSome:
            items: List[Some]
            count: int
    """
    status_code, item_schema = _find_collection_response(op)

    # Switching schema to Output schema
    try:
        new_name = f"Paged{item_schema.__name__}"
    except AttributeError:
        new_name = f"Paged{str(item_schema).replace('.', '_')}"  # typing.Any case

    new_schema = type(
        new_name,
        (paginator.Output,),
        {
            "__annotations__": {paginator.items_attribute: List[item_schema]},  # type: ignore
        },
    )  # typing: ignore

    response = op._create_response_model(new_schema)

    # Changing response model to newly created one
    op.response_models[status_code] = response


def _find_collection_response(op: Operation) -> Tuple[int, Any]:
    """
    Walks through defined operation responses and finds the first
    that is of a collection type (e.g. List[SomeSchema])
    """
    for code, resp_model in op.response_models.items():
        if resp_model is None or resp_model is NOT_SET:
            continue

        model = resp_model.__annotations__["response"]
        if is_collection_type(model):
            item_schema = get_collection_args(model)[0]
            return code, item_schema

    raise ConfigError(
        f'"{op.view_func}" has no collection response (e.g. response=List[SomeSchema])'
    )



from datetime import datetime
from enum import IntEnum

from pydantic import Field

from ninja import NinjaAPI, Query, Schema
from ninja.testing import TestClient


class Range(IntEnum):
    TWENTY = 20
    FIFTY = 50
    TWO_HUNDRED = 200


class Filter(Schema):
    to_datetime: datetime = Field(alias="to")
    from_datetime: datetime = Field(alias="from")
    range: Range = Range.TWENTY


class Data(Schema):
    an_int: int = Field(alias="int", default=0)
    a_float: float = Field(alias="float", default=1.5)


api = NinjaAPI()


@api.get("/test")
def query_params_schema(request, filters: Filter = Query(...)):
    return filters.dict()


@api.get("/test-mixed")
def query_params_mixed_schema(
    request,
    query1: int,
    query2: int = 5,
    filters: Filter = Query(...),
    data: Data = Query(...),
):
    return dict(query1=query1, query2=query2, filters=filters.dict(), data=data.dict())


# def test_request():
#     client = TestClient(api)
#     response = client.get("/test?from=1&to=2&range=20&foo=1&range2=50")
#     print("!", response.json())
#     assert response.json() == {
#         "to_datetime": "1970-01-01T00:00:02Z",
#         "from_datetime": "1970-01-01T00:00:01Z",
#         "range": 20,
#     }

#     response = client.get("/test?from=1&to=2&range=21")
#     assert response.status_code == 422


# def test_request_mixed():
#     client = TestClient(api)
#     response = client.get(
#         "/test-mixed?from=1&to=2&range=20&foo=1&range2=50&query1=2&int=3&float=1.6"
#     )
#     print(response.json())
#     assert response.json() == {
#         "data": {"a_float": 1.6, "an_int": 3},
#         "filters": {
#             "from_datetime": "1970-01-01T00:00:01Z",
#             "range": 20,
#             "to_datetime": "1970-01-01T00:00:02Z",
#         },
#         "query1": 2,
#         "query2": 5,
#     }

#     response = client.get(
#         "/test-mixed?from=1&to=2&range=20&foo=1&range2=50&query1=2&query2=10"
#     )
#     print(response.json())
#     assert response.json() == {
#         "data": {"a_float": 1.5, "an_int": 0},
#         "filters": {
#             "from_datetime": "1970-01-01T00:00:01Z",
#             "range": 20,
#             "to_datetime": "1970-01-01T00:00:02Z",
#         },
#         "query1": 2,
#         "query2": 10,
#     }

#     response = client.get("/test-mixed?from=1&to=2")
#     assert response.status_code == 422


def test_schema():
    schema = api.get_openapi_schema()
    params = schema["paths"]["/api/test"]["get"]["parameters"]
    print(params)
    assert params == [
        {
            "in": "query",
            "name": "to",
            "schema": {"format": "date-time", "title": "To", "type": "string"},
            "required": True,
        },
        {
            "in": "query",
            "name": "from",
            "schema": {"format": "date-time", "title": "From", "type": "string"},
            "required": True,
        },
        {
            "in": "query",
            "name": "range",
            "schema": {
                "allOf": [{"enum": [20, 50, 200], "title": "Range", "type": "integer"}],
                "default": 20,
            },
            "required": False,
        },
    ]


def test_schema_all_of_no_ref():
    details = {
        "default": 1,
        "allOf": [
            {"$ref": "#/components/schemas/Type"},
            {"no-ref-here": "xyzzy"},
        ],
    }
    definitions = {"Type": {"title": "Best Type Ever!"}}

    from ninja.openapi.schema import resolve_allOf

    resolve_allOf(details, definitions)

    assert details == {
        "default": 1,
        "allOf": [{"title": "Best Type Ever!"}, {"no-ref-here": "xyzzy"}],
    }



import sys

sys.stderr.write(
    """
===============================
Unsupported installation method
===============================
django-ninja does not install with `python setup.py install`.
Please use `python -m pip install` instead.
"""
)
sys.exit(1)


# The below code will never execute, however GitHub is particularly
# picky about where it finds Python packaging metadata.
# See: https://github.com/github/feedback/discussions/6456
#
# To be removed once GitHub catches up.

setup(
    name="django-ninja",
    install_requires=[
        "Django>=2.0.13",
        "pydantic>=2.0.0",
    ],
)



from typing import TYPE_CHECKING, Any, NoReturn

from django.http import Http404, HttpRequest, HttpResponse

from ninja.openapi.docs import DocsBase
from ninja.responses import Response

if TYPE_CHECKING:
    # if anyone knows a cleaner way to make mypy happy - welcome
    from ninja import NinjaAPI  # pragma: no cover


def default_home(request: HttpRequest, api: "NinjaAPI", **kwargs: Any) -> NoReturn:
    "This view is mainly needed to determine the full path for API operations"
    docs_url = f"{request.path}{api.docs_url}".replace("//", "/")
    raise Http404(f"docs_url = {docs_url}")


def openapi_json(request: HttpRequest, api: "NinjaAPI", **kwargs: Any) -> HttpResponse:
    schema = api.get_openapi_schema(path_params=kwargs)
    return Response(schema)


def openapi_view(request: HttpRequest, api: "NinjaAPI", **kwargs: Any) -> HttpResponse:
    docs: DocsBase = api.docs
    return docs.render_page(request, api, **kwargs)



import asyncio

import django
import pytest

from ninja import NinjaAPI
from ninja.security import APIKeyQuery
from ninja.testing import TestAsyncClient


@pytest.mark.skipif(django.VERSION < (3, 1), reason="requires django 3.1 or higher")
@pytest.mark.asyncio
async def test_asyncio_operations():
    api = NinjaAPI()

    class KeyQuery(APIKeyQuery):
        def authenticate(self, request, key):
            if key == "secret":
                return key

    @api.get("/async", auth=KeyQuery())
    async def async_view(request, payload: int):
        await asyncio.sleep(0)
        return {"async": True}

    @api.post("/async")
    def sync_post_to_async_view(request):
        return {"sync": True}

    client = TestAsyncClient(api)

    # Actual tests --------------------------------------------------

    # without auth:
    res = await client.get("/async?payload=1")
    assert res.status_code == 401

    # async successful
    res = await client.get("/async?payload=1&key=secret")
    assert res.json() == {"async": True}

    # async innvalid input
    res = await client.get("/async?payload=str&key=secret")
    assert res.status_code == 422

    # async call to sync method for path that have async operations
    res = await client.post("/async")
    assert res.json() == {"sync": True}

    # invalid method
    res = await client.put("/async")
    assert res.status_code == 405



from unittest import mock

import pytest

from ninja import NinjaAPI, Router
from ninja.errors import ConfigError

api = NinjaAPI()
router = Router()


@api.get("/global")
def global_op(request):
    pass


@router.get("/router")
def router_op(request):
    pass


api.add_router("/", router)


def test_api_instance():
    assert len(api._routers) == 2  # default + extra
    for _path, rtr in api._routers:
        for path_ops in rtr.path_operations.values():
            for op in path_ops.operations:
                assert op.api is api


def test_reuse_router_error():
    test_api = NinjaAPI()
    test_router = Router()
    test_api.add_router("/", test_router)

    # django debug server can attempt to import the urls twice when errors exist
    # verify we get the correct error reported
    match = "Router@'/another-path' has already been attached to API NinjaAPI:1.0.0"
    with pytest.raises(ConfigError, match=match):
        with mock.patch("ninja.main._imported_while_running_in_debug_server", False):
            test_api.add_router("/another-path", test_router)

    # The error should be ignored under debug server to allow other errors to be reported
    with mock.patch("ninja.main._imported_while_running_in_debug_server", True):
        test_api.add_router("/another-path", test_router)



from functools import wraps
from unittest import mock

import pytest

from ninja import Router
from ninja.testing import TestClient

router = Router()
client = TestClient(router)


def a_good_test_wrapper(f):
    """Validate that decorators using functools.wraps(), work as expected"""

    @wraps(f)
    def wrapper(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper


def a_bad_test_wrapper(f):
    """Validate that decorators failing to using functools.wraps(), fail"""

    def wrapper(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper


@router.get("/text")
@a_good_test_wrapper
def get_text(
    request,
):
    return "Hello World"


@router.get("/path/{item_id}")
@a_good_test_wrapper
def get_id(request, item_id):
    return item_id


@router.get("/query")
@a_good_test_wrapper
def get_query_type(request, query: int):
    return f"foo bar {query}"


@router.get("/path-query/{item_id}")
@a_good_test_wrapper
def get_query_id(request, item_id, query: int):
    return f"foo bar {item_id} {query}"


@router.get("/text-bad")
@a_bad_test_wrapper
def get_text_bad(request):
    return "Hello World"


with mock.patch("ninja.signature.details.warnings.warn_explicit"):

    @router.get("/path-bad/{item_id}")
    @a_bad_test_wrapper
    def get_id_bad(request, item_id):
        return item_id


@router.get("/query-bad")
@a_bad_test_wrapper
def get_query_type_bad(request, query: int):
    return f"foo bar {query}"


with mock.patch("ninja.signature.details.warnings.warn_explicit"):

    @router.get("/path-query-bad/{item_id}")
    @a_bad_test_wrapper
    def get_query_id_bad(request, item_id, query: int):
        return f"foo bar {item_id} {query}"


@pytest.mark.parametrize(
    "path,expected_status,expected_response",
    [
        ("/text", 200, "Hello World"),
        ("/path/id", 200, "id"),
        ("/query?query=1", 200, "foo bar 1"),
        ("/path-query/id?query=2", 200, "foo bar id 2"),
        ("/text-bad", 200, "Hello World"),  # no params so passes
        ("/path-bad/id", None, TypeError),
        ("/query-bad?query=1", None, TypeError),
        ("/path-query-bad/id?query=2", None, TypeError),
    ],
)
def test_get_path(path, expected_status, expected_response):
    if isinstance(expected_response, str):
        response = client.get(path)
        assert response.status_code == expected_status
        assert response.json() == expected_response
    else:
        match = r"Did you fail to use functools.wraps\(\) in a decorator\?"
        with pytest.raises(expected_response, match=match):
            client.get(path)



from typing import Callable

import django
from django.http import HttpRequest

try:
    from django.utils.datastructures import CaseInsensitiveMapping
except ImportError:
    from .datastructures import CaseInsensitiveMapping


# HttpHeaders copypasted from django 3.0 codebase
class HttpHeaders(CaseInsensitiveMapping):
    HTTP_PREFIX = "HTTP_"
    # PEP 333 gives two headers which aren't prepended with HTTP_.
    UNPREFIXED_HEADERS = {"CONTENT_TYPE", "CONTENT_LENGTH"}

    def __init__(self, environ):
        headers = {}
        for header, value in environ.items():
            name = self.parse_header_name(header)
            if name:
                headers[name] = value
        super().__init__(headers)

    def __getitem__(self, key):
        """Allow header lookup using underscores in place of hyphens."""
        return super().__getitem__(key.replace("_", "-"))

    @classmethod
    def parse_header_name(cls, header):
        if header.startswith(cls.HTTP_PREFIX):
            start = len(cls.HTTP_PREFIX)
            header = header[start:]
        elif header not in cls.UNPREFIXED_HEADERS:
            return None
        return header.replace("_", "-").title()


def get_headers_old(request):
    return HttpHeaders(request.META)


def get_headers_v3(request):
    return request.headers


get_headers: Callable[[HttpRequest], HttpHeaders]

if django.VERSION[0] < 3:
    get_headers = get_headers_old
else:
    get_headers = get_headers_v3



from ninja.security.apikey import APIKeyCookie, APIKeyHeader, APIKeyQuery
from ninja.security.http import HttpBasicAuth, HttpBearer
from ninja.security.session import SessionAuth, SessionAuthSuperUser

__all__ = [
    "APIKeyCookie",
    "APIKeyHeader",
    "APIKeyQuery",
    "HttpBasicAuth",
    "HttpBearer",
    "SessionAuth",
    "SessionAuthSuperUser",
    "django_auth",
]

django_auth = SessionAuth()
django_auth_superuser = SessionAuthSuperUser()



from ninja.security import APIKeyHeader


class ApiKey(APIKeyHeader):
    param_name = "X-API-Key"

    def authenticate(self, request, key):
        if key == "supersecret":
            return key


header_key = ApiKey()


@api.get("/headerkey", auth=header_key)
def apikey(request):
    return f"Token = {request.auth}"



from datetime import datetime
from http import HTTPStatus
from unittest import mock

import pytest
from django.utils import timezone

from ninja import Router
from ninja.schema import Schema
from ninja.testing import TestClient

router = Router()


@router.get("/request/build_absolute_uri")
def request_build_absolute_uri(request):
    return request.build_absolute_uri()


@router.get("/request/build_absolute_uri/location")
def request_build_absolute_uri_location(request):
    return request.build_absolute_uri("location")


@router.get("/test")
def simple_get(request):
    return "test"


client = TestClient(router)


@pytest.mark.parametrize(
    "path,expected_status,expected_response",
    [
        ("/request/build_absolute_uri", HTTPStatus.OK, "http://testlocation/"),
        (
            "/request/build_absolute_uri/location",
            HTTPStatus.OK,
            "http://testlocation/location",
        ),
    ],
)
def test_sync_build_absolute_uri(path, expected_status, expected_response):
    response = client.get(path)

    assert response.status_code == expected_status
    assert response.json() == expected_response


@pytest.mark.parametrize(
    "version, has_headers",
    [
        ((2, 0), False),
        ((2, 1), False),
        ((2, 2), True),
        ((3, 0), True),
    ],
)
def test_django_2_2_plus_headers(version, has_headers):
    with mock.patch("ninja.testing.client.django", VERSION=version):
        with mock.patch.object(client, "_call") as call:
            client.get("/test")
            request = call.call_args[0][1]
            # for Django >= 2.2 we apply a HttpHeaders instance to .headers
            assert isinstance(request.headers, mock.Mock) != has_headers


class ClientTestSchema(Schema):
    time: datetime


def test_schema_as_data():
    schema_instance = ClientTestSchema(time=timezone.now().replace(microsecond=0))

    with mock.patch.object(client, "_call") as call:
        client.post("/test", json=schema_instance)
        request = call.call_args[0][1]
        assert (
            ClientTestSchema.model_validate_json(request.body).model_dump_json()
            == schema_instance.model_dump_json()
        )


def test_json_as_body():
    schema_instance = ClientTestSchema(time=timezone.now().replace(microsecond=0))

    with mock.patch.object(client, "_call") as call:
        client.post(
            "/test",
            data=schema_instance.model_dump_json(),
            content_type="application/json",
        )
        request = call.call_args[0][1]
        assert (
            ClientTestSchema.model_validate_json(request.body).model_dump_json()
            == schema_instance.model_dump_json()
        )



import pytest
from django.db import models

from ninja import ModelSchema
from ninja.errors import ConfigError


def test_simple():
    class User(models.Model):
        firstname = models.CharField()
        lastname = models.CharField(blank=True, null=True)

        class Meta:
            app_label = "tests"

    class SampleSchema(ModelSchema):
        class Config:
            model = User
            model_fields = ["firstname", "lastname"]

        def hello(self):
            return f"Hello({self.firstname})"

    assert SampleSchema.json_schema() == {
        "title": "SampleSchema",
        "type": "object",
        "properties": {
            "firstname": {"title": "Firstname", "type": "string"},
            "lastname": {
                "anyOf": [{"type": "string"}, {"type": "null"}],
                "title": "Lastname",
            },
        },
        "required": ["firstname"],
    }

    assert SampleSchema(firstname="ninja", lastname="Django").hello() == "Hello(ninja)"

    # checking exclude ----------------------------------------------
    class SampleSchema2(ModelSchema):
        class Meta:
            model = User
            exclude = ["lastname"]

    assert SampleSchema2.json_schema() == {
        "title": "SampleSchema2",
        "type": "object",
        "properties": {
            "id": {"anyOf": [{"type": "integer"}, {"type": "null"}], "title": "ID"},
            "firstname": {"title": "Firstname", "type": "string"},
        },
        "required": ["firstname"],
    }


def test_custom():
    class CustomModel(models.Model):
        f1 = models.CharField()
        f2 = models.CharField(blank=True, null=True)

        class Meta:
            app_label = "tests"

    class CustomSchema(ModelSchema):
        f3: int
        f4: int = 1
        _private: str = "<secret>"  # private should be ignored

        class Meta:
            model = CustomModel
            fields = ["f1", "f2"]

    assert CustomSchema.json_schema() == {
        "title": "CustomSchema",
        "type": "object",
        "properties": {
            "f1": {"title": "F1", "type": "string"},
            "f2": {"anyOf": [{"type": "string"}, {"type": "null"}], "title": "F2"},
            "f3": {"title": "F3", "type": "integer"},
            "f4": {"title": "F4", "default": 1, "type": "integer"},
        },
        "required": ["f3", "f1"],
    }


def test_config():
    class Category(models.Model):
        title = models.CharField()

        class Meta:
            app_label = "tests"

    with pytest.raises(ConfigError):

        class CategorySchema(ModelSchema):
            class Meta:
                model = Category


def test_optional():
    class OptModel(models.Model):
        title = models.CharField()
        other = models.CharField(null=True)

        class Meta:
            app_label = "tests"

    class OptSchema(ModelSchema):
        class Meta:
            model = OptModel
            fields = "__all__"
            fields_optional = ["title"]

    class OptSchema2(ModelSchema):
        class Meta:
            model = OptModel
            fields = "__all__"
            fields_optional = "__all__"

    assert OptSchema.json_schema().get("required") is None

    assert OptSchema2.json_schema().get("required") is None


def test_fields_all():
    class SomeModel(models.Model):
        field1 = models.CharField()
        field2 = models.CharField(blank=True, null=True)

        class Meta:
            app_label = "tests"

    class SomeSchema(ModelSchema):
        class Meta:
            model = SomeModel
            fields = "__all__"

    print(SomeSchema.json_schema())
    assert SomeSchema.json_schema() == {
        "title": "SomeSchema",
        "type": "object",
        "properties": {
            "id": {"anyOf": [{"type": "integer"}, {"type": "null"}], "title": "ID"},
            "field1": {"title": "Field1", "type": "string"},
            "field2": {
                "anyOf": [{"type": "string"}, {"type": "null"}],
                "title": "Field2",
            },
        },
        "required": ["field1"],
    }


def test_model_schema_without_config():
    with pytest.raises(
        ConfigError,
        match=r"ModelSchema class 'NoConfigSchema' requires a 'Meta' \(or a 'Config'\) subclass",
    ):

        class NoConfigSchema(ModelSchema):
            x: int
